"""
POC Enrollment operations for use by Coach APIs.

Does not include any access control, be sure to check access before calling.
"""

from courseware.courses import get_course_about_section
from courseware.courses import get_course_by_id
from django.contrib.auth.models import User
from django.conf import settings
from django.core.urlresolvers import reverse
from django.core.mail import send_mail
from edxmako.shortcuts import render_to_string
from microsite_configuration import microsite
from xmodule.modulestore.django import modulestore
from xmodule.error_module import ErrorDescriptor

from .models import (
    PocMembership,
    PocFutureMembership,
)
from .overrides import get_current_poc


class EmailEnrollmentState(object):
    """ Store the complete enrollment state of an email in a class """
    def __init__(self, poc, email):
        exists_user = User.objects.filter(email=email).exists()
        if exists_user:
            user = User.objects.get(email=email)
            poc_member = PocMembership.objects.filter(poc=poc, student=user)
            in_poc = poc_member.exists()
            full_name = user.profile.name
        else:
            user = None
            in_poc = False
            full_name = None
        self.user = exists_user
        self.member = user
        self.full_name = full_name
        self.in_poc = in_poc

    def __repr__(self):
        return "{}(user={}, member={}, in_poc={}".format(
            self.__class__.__name__,
            self.user,
            self.member,
            self.in_poc,
        )

    def to_dict(self):
        return {
            'user': self.user,
            'member': self.member,
            'in_poc': self.in_poc,
        }


def enroll_email(poc, student_email, auto_enroll=False, email_students=False, email_params=None):
    if email_params is None:
        email_params = get_email_params(poc, True)
    previous_state = EmailEnrollmentState(poc, student_email)

    if previous_state.user:
        if not previous_state.in_poc:
            user = User.objects.get(email=student_email)
            membership = PocMembership(
                poc=poc, student=user, active=True
            )
            membership.save()
        elif auto_enroll:
            # activate existing memberships
            membership = PocMembership.objects.get(student=user, poc=poc)
            membership.active = True
            membership.save()
        if email_students:
            email_params['message'] = 'enrolled_enroll'
            email_params['email_address'] = student_email
            email_params['full_name'] = previous_state.full_name
            send_mail_to_student(student_email, email_params)
    else:
        membership = PocFutureMembership(
            poc=poc, auto_enroll=auto_enroll, email=student_email
        )
        membership.save()
        if email_students:
            email_params['message'] = 'allowed_enroll'
            email_params['email_address'] = student_email
            send_mail_to_student(student_email, email_params)

    after_state = EmailEnrollmentState(poc, student_email)

    return previous_state, after_state


def unenroll_email(poc, student_email, email_students=False, email_params=None):
    if email_params is None:
        email_params = get_email_params(poc, True)
    previous_state = EmailEnrollmentState(poc, student_email)

    if previous_state.in_poc:
        PocMembership.objects.get(
            poc=poc, student=previous_state.member
        ).delete()
        if email_students:
            email_params['message'] = 'enrolled_unenroll'
            email_params['email_address'] = student_email
            email_params['full_name'] = previous_state.full_name
            send_mail_to_student(student_email, email_params)
    else:
        if PocFutureMembership.objects.filter(
            poc=poc, email=student_email
        ).exists():
            PocFutureMembership.objects.get(
                poc=poc, email=student_email
            ).delete()
        if email_students:
            email_params['message'] = 'allowed_unenroll'
            email_params['email_address'] = student_email
            send_mail_to_student(student_email, email_params)

    after_state = EmailEnrollmentState(poc, student_email)

    return previous_state, after_state


def get_email_params(poc, auto_enroll, secure=True):
    protocol = 'https' if secure else 'http'
    course_id = poc.course_id

    stripped_site_name = microsite.get_value(
        'SITE_NAME',
        settings.SITE_NAME
    )
    registration_url = u'{proto}://{site}{path}'.format(
        proto=protocol,
        site=stripped_site_name,
        path=reverse('student.views.register_user')
    )
    course_url = u'{proto}://{site}{path}'.format(
        proto=protocol,
        site=stripped_site_name,
        path=reverse(
            'course_root',
            kwargs={'course_id': course_id.to_deprecated_string()}
        )
    )

    course_about_url = None
    if not settings.FEATURES.get('ENABLE_MKTG_SITE', False):
        course_about_url = u'{proto}://{site}{path}'.format(
            proto=protocol,
            site=stripped_site_name,
            path=reverse(
                'about_course',
                kwargs={'course_id': course_id.to_deprecated_string()}
            )
        )

    email_params = {
        'site_name': stripped_site_name,
        'registration_url': registration_url,
        'course': poc,
        'auto_enroll': auto_enroll,
        'course_url': course_url,
        'course_about_url': course_about_url,
    }
    return email_params


def send_mail_to_student(student, param_dict):
    if 'course' in param_dict:
        param_dict['course_name'] = param_dict['course'].display_name

    param_dict['site_name'] = microsite.get_value(
        'SITE_NAME',
        param_dict['site_name']
    )

    subject = None
    message = None

    message_type = param_dict['message']

    email_template_dict = {
        'allowed_enroll': (
            'pocs/enroll_email_allowedsubject.txt',
            'pocs/enroll_email_allowedmessage.txt'
        ),
        'enrolled_enroll': (
            'pocs/enroll_email_enrolledsubject.txt',
            'pocs/enroll_email_enrolledmessage.txt'
        ),
        'allowed_unenroll': (
            'pocs/unenroll_email_subject.txt',
            'pocs/unenroll_email_allowedmessage.txt'
        ),
        'enrolled_unenroll': (
            'pocs/unenroll_email_subject.txt',
            'pocs/unenroll_email_enrolledmessage.txt'
        ),
    }

    subject_template, message_template = email_template_dict.get(
        message_type, (None, None)
    )
    if subject_template is not None and message_template is not None:
        subject = render_to_string(subject_template, param_dict)
        message = render_to_string(message_template, param_dict)

    if subject and message:
        message = message.strip()

        subject = ''.join(subject.splitlines())
        from_address = microsite.get_value(
            'email_from_address',
            settings.DEFAULT_FROM_EMAIL
        )

        send_mail(
            subject,
            message,
            from_address,
            [student],
            fail_silently=False
        )


def get_all_pocs_for_user(user):
    """return all POCS to which the user is registered

    Returns a list of dicts: {
        poc_name: <formatted title of POC course>
        poc_url: <url to view this POC>
        poc_active: True if this poc is currently the 'active' one
        mooc_name: <formatted title of the MOOC course for this POC>
        mooc_url: <url to view this MOOC>
    }
    """
    current_active_poc = get_current_poc(user)
    if user.is_anonymous():
        return []
    memberships = []
    for membership in PocMembership.memberships_for_user(user):
        course = get_course_by_id(membership.poc.course_id)
        poc = membership.poc
        poc_title = poc.display_name
        mooc_title = get_course_about_section(course, 'title')
        url = reverse(
            'switch_active_poc',
            args=[course.id.to_deprecated_string(), membership.poc.id]
        )
        mooc_url = reverse(
            'switch_active_poc',
            args=[course.id.to_deprecated_string(),]
        )
        memberships.append({
            'poc_name': poc_title,
            'poc_url': url,
            'active': membership.poc == current_active_poc,
            'mooc_name': mooc_title,
            'mooc_url': mooc_url,
        })
    return memberships

def get_poc_membership_triplets(user, course_org_filter, org_filter_out_set):
    """
    Get the relevant set of (PersonalOnlineCourse, PocMembership, Course)
    triplets to be displayed on a student's dashboard.
    """
    # only active memberships for now
    for membership in PocMembership.memberships_for_user(user):
        poc = membership.poc
        store = modulestore()
        with store.bulk_operations(poc.course_id):
            course = store.get_course(poc.course_id)
            if course and not isinstance(course, ErrorDescriptor):
                # if we are in a Microsite, then filter out anything that is not
                # attributed (by ORG) to that Microsite
                if course_org_filter and course_org_filter != course.location.org:
                    continue
                # Conversely, if we are not in a Microsite, then let's filter out any enrollments
                # with courses attributed (by ORG) to Microsites
                elif course.location.org in org_filter_out_set:
                    continue

                yield (poc, membership, course)
            else:
                log.error("User {0} enrolled in {2} course {1}".format(
                    user.username, poc.course_id, "broken" if course else "non-existent"
                ))
